//
// the main function for "fit", a least-squares
// curve-fitting program based on the Marquardt algorithm.
//

#include "fit.h"

int main(int argc, char** argv)
{
    misc mi;                 // struct of miscellaneous quantities
    misc* mp = &mi;
    flags f = {0};           // flags for output, precision, quiet,
                             // sigma, eps_test, gam_eps_test
    flags* fp = &f;
    array ar;

// get options and output filename if any

    get_opts(argc, argv, fp, mp);

// arguments must include datafile and paramfile

    if (argc < 3) {
        prn_usage(argv);
        exit(1);
    }

// get npar, conf, maxiter, lambda, and parameter names
  
     get_func_info(mp);

// get number of data points from datafile

     int npts = get_npts(argv);

// allocate needed memory

    dbl*   x = ar.alloc_d(npts);
    dbl*   y = ar.alloc_d(npts);
    dbl*   s = ar.alloc_d(npts);
    dbl*   r = ar.alloc_d(npts);
    dbl*   b = ar.alloc_d(mi.npar);
    dbl*  bs = ar.alloc_d(mi.npar);
    dbl*  db = ar.alloc_d(mi.npar);
    dbl*   p = ar.alloc_d(mi.npar);
    dbl*   g = ar.alloc_d(mi.npar);
    dbl*  sf = ar.alloc_d(mi.npar);
    dbl**  a = ar.alloc_d(mi.npar, mi.npar);
    dbl** sa = ar.alloc_d(mi.npar, mi.npar);

// get data into x, y, and s
    get_data(argv, x, y, s);

// get initial parameter values into b
    get_params(argv, b, mp);


    resid(x, y, r, b, npts);
    mi.phi = sumsq(r, s, npts);

// output initial parameter values and initial phi

    if (!f.q) output(b, a, mp, fp, npts, 0);


// start iterations

    for (mi.iter_no = 1 ; mi.iter_no < mi.maxiter ; mi.iter_no++) {
        dbl phi_old = mi.phi;
        new_a(x, s, r, b, p, g, a, mp, npts);
        if (f.et) {
            scale(a, g, sf, mi.npar);
            if (gjr(a, mi.npar)) {
                err("on attempted final output, the a-matrix\n"
                    "is found to be singular.");
            }
            calc_db_gamma(db, a, g, mp);
            new_b(db, b, sf, mi.npar);
            resid(x, y, r, b, npts);
            mi.phi = sumsq(r, s, npts);
            new_a(x, s, r, b, p, g, a, mp, npts);
            if (f.get) output(b, a, mp, fp, npts, 4);
            else output(b, a, mp, fp, npts, 3);
            if (f.o) mk_outfile(x, y, s, b, mp, fp, npts);
            exit(0);
        }
        copy_mat(sa, a, mi.npar);      // store the "a" matrix in sa

// scale a, add lambda to its diagonal elements,
// invert the resulting matrix, and find correction db

        for (int m = 0 ; ; m++) {
            copy_mat(a, sa, mi.npar);  // get a from storage
            scale(a, g, sf, mi.npar);
            for (int j = 0 ; j < mi.npar ; j++)
                a[j][j] += mi.lambda;
            if (!gjr(a, mi.npar))      // if a is not singular
                break;                 // go on to find db
            mi.lambda *= 10.0;         // else increase lambda
            if (m > 3) {               // give up after 4 tries
                err("found a singular a-matrix.  maybe\n   "
                    "your chosen parameters are too highly\n   "
                    "correlated, or maybe you could start\n   "
                    "with a larger value of lambda.");
            }
        }

// compute gamma and db from a and g

        calc_db_gamma(db, a, g, mp);

// correct the b-vector and get a new phi

        copy_vec(bs, b, mi.npar);      // store old b in bs
        new_b(db, b, sf, mi.npar);     // correct b
        resid(x, y, r, b, npts);       // resids with new b's
        mi.phi = sumsq(r, s, npts);
        if (!f.q)
            output(b, a, mp, fp, npts, 1); // intermed output

// test for non-convergence

        if (mi.phi <= phi_old) {
            if (gam_lam_test(mi.gamma, mi.lambda)) // bad news
                output(b, a, mp, fp, npts, 2);

            if (eps_test(b, db, sf, mi.npar)) {
                f.et = yes;             // passes eps test, so
                continue;               // do final "iteration"
            }

            if (mi.lambda >= 1.0e-9)    // did not pass eps test, so
                mi.lambda *= 0.1;       // reduce lambda and iterate again
        }
        else {                          // phi > phi_old
            if (mi.gamma < GAMCR) {
                while (mi.phi > phi_old) {
                    if (gam_eps_test(b, bs, db, sf, mi.npar)) {
                        f.get = yes;    // passes gam_eps_test, so
                        break;          // start final "iteration"
                    }
                    resid(x, y, r, b, npts);    // calculate
                    mi.phi = sumsq(r, s, npts); // a new phi
                }

                if (!f.get)             // phi should be < phi_old now
                    gassert(mi.phi <= phi_old);
                
                if (mi.lambda >= 1.0e-9)
                    mi.lambda *= 0.5;   // try only halving it
            }
            else {                      // gamma > GAMCR, so
                mi.lambda *= 10.0;      // increase lambda
            }
        }
    }                                   // end of iteration loop

    if (mi.iter_no < mi.maxiter)
        return 0;
    gassert(mi.iter_no == mi.maxiter);
    output(b, a, mp, fp, npts, 5);      // reached max iterations
    exit(1);
}

